import ckan.logic as logic
import ckan.plugins.toolkit as tk
import ckanext.relationship.logic.schema as schema
from ckan.logic import validate
from ckanext.relationship.model.relationship import Relationship

NotFound = logic.NotFound

_actions = {}


def action(func):
    func.__name__ = f'relationship_{func.__name__}'
    _actions[func.__name__] = func
    return func


def get_actions():
    return _actions.copy()


@action
@validate(schema.relation_create)
def relation_create(context, data_dict):
    tk.check_access('relationship_relation_create', context, data_dict)

    model = context['model']

    subject_id = data_dict['subject_id']
    object_id = data_dict['object_id']
    relation_type = data_dict.get('relation_type')

    relation = Relationship(subject_id=subject_id,
                            object_id=object_id,
                            relation_type=relation_type
                            )

    reverse_relation = Relationship(subject_id=object_id,
                                    object_id=subject_id,
                                    relation_type=Relationship.reverse_reletion_type[relation_type]
                                    )

    context['session'].add(relation)
    context['session'].add(reverse_relation)
    context['session'].commit()

    return [rel.as_dict() for rel in (relation, reverse_relation)]


@action
@validate(schema.relations_list)
def relations_list(context, data_dict):
    tk.check_access('relationship_relations_list', context, data_dict)

    subject_id = data_dict['subject_id']
    object_entity = data_dict['object_entity']
    object_entity = object_entity if object_entity != 'organization' else 'group'
    object_type = data_dict['object_type']
    relation_type = data_dict.get('relation_type')

    relations = Relationship.by_object_type(subject_id,
                                            object_entity,
                                            object_type,
                                            relation_type
                                            )
    if not relations:
        return []

    return [rel.as_dict() for rel in relations]


@action
@validate(schema.relation_delete)
def relation_delete(context, data_dict):
    tk.check_access('relationship_relation_delete', context, data_dict)

    subject_id = data_dict['subject_id']
    object_id = data_dict['object_id']
    relation_type = data_dict.get('relation_type')

    relation = (context['session'].query(Relationship)
                .filter(Relationship.subject_id == data_dict['subject_id'],
                        Relationship.object_id == data_dict['object_id'],
                        Relationship.relation_type == data_dict.get('relation_type'))
                .one_or_none()
                )

    if not relation:
        raise tk.ObjectNotFound('Relation not found')

    reverse_relation = (context['session'].query(Relationship)
                        .filter(Relationship.subject_id == data_dict['object_id'],
                                Relationship.object_id == data_dict['subject_id'],
                                Relationship.relation_type == Relationship.reverse_reletion_type[
                                    data_dict.get('relation_type')])
                        .one_or_none()
                        )

    context['session'].delete(relation)
    context['session'].delete(reverse_relation)
    context['session'].commit()

    return [rel.as_dict() for rel in (relation, reverse_relation)]


@action
@validate(schema.get_entity_list)
def get_entity_list(context, data_dict):
    tk.check_access('relationship_get_entity_list', context, data_dict)

    model = context['model']

    entity = data_dict['entity']
    entity = entity if entity != 'organization' else 'group'

    entity_type = data_dict['entity_type']

    entity_class = logic.model_name_to_class(model, entity)

    entity_list = (context['session'].query(entity_class.id, entity_class.name, entity_class.title)
                   .filter(entity_class.state != 'deleted')
                   .filter(entity_class.type == entity_type).all())

    return entity_list



import ckan.plugins.toolkit as tk

_helpers = {}


def helper(func):
    func.__name__ = f'relationship_{func.__name__}'
    _helpers[func.__name__] = func


def get_helpers():
    return _helpers.copy()


@helper
def get_entity_list(entity, entity_type, include_private=True):
    context = {}
    if entity == 'package':
        entity_list = tk.get_action('package_search')(context, {'fq': f'type:{entity_type}',
                                                                'fl': 'id, name, title',
                                                                'rows': 1000,
                                                                'include_private': include_private})
        entity_list = entity_list['results']
    else:
        entity_list = tk.get_action('relationship_get_entity_list')(context, {'entity': entity,
                                                                              'entity_type': entity_type})
        entity_list = [{'id': id, 'name': name, 'title': title} for id, name, title in entity_list]
    return entity_list


@helper
def get_current_relations_list(data, field):
    subject_id = field.get('id')
    subject_name = field.get('name')
    if not subject_id and not subject_name:
        return []
    related_entity = data['related_entity']
    related_entity_type = data['related_entity_type']
    relation_type = data['relation_type']

    current_relation_by_id = []
    current_relation_by_name = []

    if subject_id:
        current_relation_by_id = tk.get_action('relationship_relations_list')({}, {'subject_id': subject_id,
                                                                                   'object_entity': related_entity,
                                                                                   'object_type': related_entity_type,
                                                                                   'relation_type': relation_type})
    if subject_name:
        current_relation_by_name = tk.get_action('relationship_relations_list')({}, {'subject_id': subject_name,
                                                                                     'object_entity': related_entity,
                                                                                     'object_type': related_entity_type,
                                                                                     'relation_type': relation_type})
    return [rel['object_id'] for rel in current_relation_by_id + current_relation_by_name]



import ckan.model as model
import ckan.logic as logic
from ckan.model import core
from ckan.model.types import make_uuid
from sqlalchemy import Column, Text, or_, and_

from .base import Base


class Relationship(Base):
    __tablename__ = 'relationship_relationship'
    id: str = Column(Text, primary_key=True, default=make_uuid)
    subject_id: str = Column(Text, nullable=False)
    object_id: str = Column(Text, nullable=False)
    relation_type: str = Column(Text, nullable=False)

    reverse_reletion_type = {'related_to': 'related_to',
                             'child_of': 'parent_of',
                             'parent_of': 'child_of'
                             }

    def __repr__(self):
        return (
            'Relationship('
            f'id={self.id!r},'
            f'subject_id={self.subject_id!r},'
            f'object_id={self.object_id!r},'
            f'relation_type={self.relation_type!r},'
            ')'
        )

    def as_dict(self):
        id = self.id
        subject_id = self.subject_id
        object_id = self.object_id
        relation_type = self.relation_type
        return ({'id': id,
                 'subject_id': subject_id,
                 'object_id': object_id,
                 'relation_type': relation_type
                 })

    @classmethod
    def by_object_id(cls, subject_id, object_id, relation_type):
        return model.Session.query(cls). \
            filter(cls.subject_id == subject_id). \
            filter(cls.object_id == object_id). \
            filter(cls.relation_type == relation_type).one_or_none()

    @classmethod
    def by_object_type(cls, subject_id, object_entity, object_type, relation_type):
        object_class = logic.model_name_to_class(model, object_entity)

        return model.Session.query(cls). \
            filter(cls.subject_id == subject_id). \
            filter(object_class.id == cls.object_id). \
            filter(object_class.type == object_type). \
            filter(cls.relation_type == relation_type).all()



import ckan.plugins as plugins
import ckan.plugins.toolkit as tk
import ckanext.relationship.helpers as helpers
import ckanext.relationship.logic.action as action
import ckanext.relationship.logic.auth as auth
import ckanext.relationship.logic.validators as validators
import ckanext.scheming.helpers as sch
from ckan.lib.search import rebuild
from ckan.logic import NotFound


class RelationshipPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IActions)
    plugins.implements(plugins.IAuthFunctions)
    plugins.implements(plugins.IValidators)
    plugins.implements(plugins.ITemplateHelpers)
    plugins.implements(plugins.IPackageController, inherit=True)

    # IConfigurer
    def update_config(self, config_):
        tk.add_template_directory(config_, 'templates')
        tk.add_public_directory(config_, 'public')
        tk.add_resource('fanstatic', 'relationship')

    # IActions
    def get_actions(self):
        return action.get_actions()

    # IAuthFunctions
    def get_auth_functions(self):
        return auth.get_auth_functions()

    # IValidators
    def get_validators(self):
        return validators.get_validators()

    # ITemplateHelpers
    def get_helpers(self):
        return helpers.get_helpers()

    # IPackageController
    def after_create(self, context, pkg_dict):
        return _update_relations(pkg_dict)

    def after_update(self, context, pkg_dict):
        return _update_relations(pkg_dict)

    def before_index(self, pkg_dict):
        pkg_id = pkg_dict['id']
        pkg_type = pkg_dict['type']
        schema = sch.scheming_get_schema('dataset', pkg_type)
        if not schema:
            return pkg_dict
        relations_info = _get_relations_info(schema)
        for related_entity, related_entity_type, relation_type in relations_info:
            relations_list = tk.get_action('relationship_relations_list')({}, {'subject_id': pkg_id,
                                                                               'object_entity': related_entity,
                                                                               'object_type': related_entity_type,
                                                                               'relation_type': relation_type})
            if not relations_list:
                continue

            relations_ids = [rel['object_id'] for rel in relations_list]
            pkg_dict[f'vocab_{relation_type}_{related_entity_type}s'] = relations_ids

        return pkg_dict


def _get_relations_info(schema):
    return [(field['related_entity'], field['related_entity_type'], field['relation_type'])
            for field in schema['dataset_fields'] if 'relationship_related_entity' in field.get('validators', '')]


def _update_relations(pkg_dict):
    subject_id = pkg_dict['id']
    add_relations = pkg_dict.get('add_relations', [])
    del_relations = pkg_dict.get('del_relations', [])
    for object_id, relation_type in add_relations + del_relations:
        if (object_id, relation_type) in add_relations:
            tk.get_action('relationship_relation_create')({}, {'subject_id': subject_id,
                                                               'object_id': object_id,
                                                               'relation_type': relation_type
                                                               })
        else:
            tk.get_action('relationship_relation_delete')({}, {'subject_id': subject_id,
                                                               'object_id': object_id,
                                                               'relation_type': relation_type
                                                               })
        
        try:
            tk.get_action('package_show')({}, {'id': object_id})
            rebuild(object_id)
        except NotFound:
            pass
    rebuild(subject_id)
    return pkg_dict



from ckan.logic.schema import validator_args
import ckan.plugins.toolkit as tk


@validator_args
def relation_create(not_empty, one_of, ignore_missing):
    return {
        'subject_id': [
            not_empty,
        ],
        'object_id': [
            not_empty,
        ],
        'relation_type': [
            one_of(['related_to', 'child_of', 'parent_of']),
        ],
    }


@validator_args
def relations_list(not_empty, one_of, ignore_missing):
    return {
        'subject_id': [
            not_empty,
        ],
        'object_entity': [
            not_empty,
            one_of(['package', 'organization', 'group']),
        ],
        'object_type': [
            not_empty,
        ],
        'relation_type': [
            not_empty,
            one_of(['related_to', 'child_of', 'parent_of']),
        ],
    }


@validator_args
def relation_delete(not_empty, one_of, ignore_missing):
    return {
        'subject_id': [
            not_empty,
        ],
        'object_id': [
            not_empty,
        ],
        'relation_type': [
            one_of(['related_to', 'child_of', 'parent_of']),
        ],
    }


@validator_args
def get_entity_list(not_empty, one_of, ignore_missing):
    return {
        'entity': [
            not_empty,
            one_of(['package', 'organization', 'group']),
        ],
        'entity_type': [
            not_empty,
        ],
    }



'''Create relationship table

Revision ID: aca2ff1d3ce4
Revises:
Create Date: 2021-07-02 14:40:37.719003

'''
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'aca2ff1d3ce4'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'relationship_relationship',
        sa.Column('id', sa.Text, primary_key=True),
        sa.Column('subject_id', sa.Text, nullable=False),
        sa.Column('object_id', sa.Text, nullable=False),
        sa.Column('relation_type', sa.Text, nullable=False),
    )


def downgrade():
    op.drop_table('relationship_relationship')



import json

import ckan.plugins.toolkit as tk
import six
from ckanext.scheming.validation import scheming_validator, scheming_multiple_choice_output
from ckantoolkit import missing

_validators = {}


def validator(func):
    func.__name__ = f'relationship_{func.__name__}'
    _validators[func.__name__] = func


def get_validators():
    return _validators.copy()


@validator
@scheming_validator
def related_entity(field, schema):
    related_entity = field.get('related_entity')
    related_entity_type = field.get('related_entity_type')
    relation_type = field.get('relation_type')

    def validator(key, data, errors, context):
        if field.get('required') and data[key] is missing:
            errors[key].append(tk._('Select at least one'))

        entity_id = data.get(('id',))

        current_relations = _get_current_relations(entity_id, related_entity, related_entity_type, relation_type)
        selected_relations = _get_selected_relations(data[key])

        data[key] = json.dumps([value for value in selected_relations])

        add_relations = selected_relations - current_relations
        del_relations = current_relations - selected_relations

        data[('add_relations',)] = data.get(('add_relations',), [])
        data[('del_relations',)] = data.get(('del_relations',), [])

        data[('add_relations',)].extend([(rel, relation_type) for rel in add_relations])
        data[('del_relations',)].extend([(rel, relation_type) for rel in del_relations])

    return validator


def _get_current_relations(entity_id, related_entity, related_entity_type, relation_type):
    if entity_id:
        current_relations = tk.get_action('relationship_relations_list')({}, {'subject_id': entity_id,
                                                                              'object_entity': related_entity,
                                                                              'object_type': related_entity_type,
                                                                              'relation_type': relation_type})
        current_relations = [rel['object_id'] for rel in current_relations]
    else:
        current_relations = []
    return set(current_relations)


def _get_selected_relations(selected):
    selected_relations = selected
    if selected_relations is not missing:
        selected_relations = scheming_multiple_choice_output(selected)
        selected_relations = [] if selected_relations == [''] else selected_relations
    else:
        selected_relations = []
    return set(selected_relations)



from ckan.plugins import toolkit as tk


_auth = {}


def auth(func):
    func.__name__ = f'relationship_{func.__name__}'
    _auth[func.__name__] = func
    return func


def get_auth_functions():
    return _auth.copy()


@auth
def relation_create(context, data_dict):
    return {'success': True}


@auth
@tk.auth_allow_anonymous_access
def relations_list(context, data_dict):
    return {'success': True}


@auth
def relation_delete(context, data_dict):
    return {'success': True}

@auth
@tk.auth_allow_anonymous_access
def get_entity_list(context, data_dict):
    return {'success': True}
